@@ -1,16 +1,15 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2015 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2021 -- leonerd@leonerd.org.uk
 
-package Parser::MGC;
+package Parser::MGC 0.17;
 
-use strict;
+use v5.14;
 use warnings;
 
-our $VERSION = '0.14';
-
 use Carp;
+use Feature::Compat::Try;
 
 use File::Slurp::Tiny qw( read_file );
 use Scalar::Util qw( blessed );
@@ -317,6 +316,32 @@ sub pos
    return pos $self->{str};
 }
 
+=head2 take
+
+   $str = $parser->take( $len )
+
+I<Since version 0.16.>
+
+Returns the next C<$len> characters directly from the input, prior to any
+whitespace or comment skipping. This does I<not> take account of any
+end-of-scope marker that may be pending. It is intended for use by parsers of
+partially-binary protocols, or other situations in which it would be incorrect
+for the end-of-scope marker to take effect at this time.
+
+=cut
+
+sub take
+{
+   my $self = shift;
+   my ( $len ) = @_;
+
+   my $start = pos( $self->{str} );
+
+   pos( $self->{str} ) += $len;
+
+   return substr( $self->{str}, $start, $len );
+}
+
 =head2 where
 
    ( $lineno, $col, $text ) = $parser->where
@@ -449,8 +474,9 @@ which will be passed the actual C<$parser> object as its first argument.
 
    $ret = $parser->maybe( $code )
 
-Attempts to execute the given C<$code> reference in scalar context, and
-returns what it returned, accepting that it might fail.
+Attempts to execute the given C<$code> in scalar context, and returns what it
+returned, accepting that it might fail. C<$code> may either be a CODE
+reference or a method name given as a string.
 
 If the code fails (either by calling C<fail> itself, or by propagating a
 failure from another method it invoked) before it has invoked C<commit>, then
@@ -487,14 +513,15 @@ sub maybe
    my $committed = 0;
    local $self->{committer} = sub { $committed++ };
 
-   my $ret;
-   eval { $ret = $code->( $self ); 1 } and return $ret;
-   my $e = $@;
+   try {
+      return $self->$code;
+   }
+   catch ( $e ) {
+      pos($self->{str}) = $pos;
 
-   pos($self->{str}) = $pos;
-
-   die $e if $committed or not _isa_failure( $e );
-   return undef;
+      die $e if $committed or not _isa_failure( $e );
+      return undef;
+   }
 }
 
 =head2 scope_of
@@ -502,8 +529,9 @@ sub maybe
    $ret = $parser->scope_of( $start, $code, $stop )
 
 Expects to find the C<$start> pattern, then attempts to execute the given
-C<$code> reference, then expects to find the C<$stop> pattern. Returns
-whatever the code reference returned.
+C<$code>, then expects to find the C<$stop> pattern. Returns whatever the
+code returned. C<$code> may either be a CODE reference of a method name given
+as a string.
 
 While the code is being executed, the C<$stop> pattern will be used by the
 token parsing methods as an end-of-scope marker; causing them to raise a
@@ -538,22 +566,49 @@ immediately propagated to the caller.
 sub scope_of
 {
    my $self = shift;
-   my ( $start, $code, $stop ) = @_;
+   $self->_scope_of( 0, @_ );
+}
+
+sub _scope_of
+{
+   my $self = shift;
+   my ( $commit_if_started, $start, $code, $stop ) = @_;
 
    ref $stop or $stop = qr/\Q$stop/;
 
    $self->expect( $start ) if defined $start;
 
+   $self->commit if $commit_if_started;
+
    local $self->{endofscope} = $stop;
    local $self->{scope_level} = $self->{scope_level} + 1;
 
-   my $ret = $code->( $self );
+   my $ret = $self->$code;
 
    $self->expect( $stop );
 
    return $ret;
 }
 
+=head2 committed_scope_of
+
+   $ret = $parser->committed_scope_of( $start, $code, $stop )
+
+I<Since version 0.16.>
+
+A variant of L</scope_of> that calls L</commit> after a successful match of
+the start pattern. This is usually what you want if using C<scope_of> from
+within an C<any_of> choice, if no other alternative following this one could
+possibly match if the start pattern has.
+
+=cut
+
+sub committed_scope_of
+{
+   my $self = shift;
+   $self->_scope_of( 1, @_ );
+}
+
 =head2 list_of
 
    $ret = $parser->list_of( $sep, $code )
@@ -561,7 +616,8 @@ sub scope_of
 Expects to find a list of instances of something parsed by C<$code>,
 separated by the C<$sep> pattern. Returns an ARRAY ref containing a list of
 the return values from the C<$code>. A single trailing delimiter is allowed,
-and does not affect the return value.
+and does not affect the return value. C<$code> may either be a CODE reference
+or a method name given as a string.
 
 This method does not consider it an error if the returned list is empty; that
 is, that the scope ended before any item instances were parsed from it.
@@ -600,13 +656,16 @@ sub list_of
       $committed = 0;
       my $pos = pos $self->{str};
 
-      eval { push @ret, $code->( $self ); 1 } and next;
-      my $e = $@;
+      try {
+         push @ret, $self->$code;
+         next;
+      }
+      catch ( $e ) {
+         pos($self->{str}) = $pos;
+         die $e if $committed or not _isa_failure( $e );
 
-      pos($self->{str}) = $pos;
-      die $e if $committed or not _isa_failure( $e );
-
-      last;
+         last;
+      }
    }
    continue {
       if( defined $sep ) {
@@ -654,9 +713,11 @@ sub sequence_of
 
 I<Since version 0.06.>
 
-Expects that one of the given code references can parse something from the
-input, returning what it returned. Each code reference may indicate a failure
+Expects that one of the given code instances can parse something from the
+input, returning what it returned. Each code instance may indicate a failure
 to parse by calling the C<fail> method or otherwise propagating a failure.
+Each code instance may either be a CODE reference or a method name given as a
+string.
 
 This may be considered to be similar to the C<|> regexp operator for forming
 alternations of possible parse trees.
@@ -688,26 +749,27 @@ sub any_of
    my $self = shift;
 
    while( @_ ) {
+      my $code = shift;
       my $pos = pos $self->{str};
 
       my $committed = 0;
       local $self->{committer} = sub { $committed++ };
 
-      my $ret;
-      eval { $ret = shift->( $self ); 1 } and return $ret;
-      my $e = $@;
+      try {
+         return $self->$code;
+      }
+      catch ( $e ) {
+         pos( $self->{str} ) = $pos;
 
-      pos( $self->{str} ) = $pos;
-
-      die $e if $committed or not _isa_failure( $e );
+         die $e if $committed or not _isa_failure( $e );
+      }
    }
 
    $self->fail( "Found nothing parseable" );
 }
 
 sub one_of {
-   carp "Parser::MGC->one_of is deprecated; use ->any_of instead";
-   goto &any_of;
+   croak "Parser::MGC->one_of is deprecated; use ->any_of instead";
 }
 
 =head2 commit
@@ -738,6 +800,8 @@ rather than by attempting other alternatives.
     );
  }
 
+Though in this common pattern, L</committed_scope_of> may be used instead.
+
 =cut
 
 sub commit
@@ -909,8 +973,7 @@ sub substring_before
       $end = length $self->{str};
    }
 
-   pos( $self->{str} ) = $end;
-   return substr( $self->{str}, $start, $end - $start );
+   return $self->take( $end - $start );
 }
 
 =head2 generic_token
